;;; -*- Mode: Lisp; -*-

;;;  (c) 2006 David Lichteblau (david@lichteblau.com)
;;; based on the null backend by:
;;;  (c) 2005 Christophe Rhodes (c.rhodes@gold.ac.uk)

;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Library General Public
;;; License as published by the Free Software Foundation; either
;;; version 2 of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Library General Public License for more details.
;;;
;;; You should have received a copy of the GNU Library General Public
;;; License along with this library; if not, write to the 
;;; Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
;;; Boston, MA  02111-1307  USA.

(in-package :clim-gtkairo)

;;; Locking rule for this file: WITH-GTK is used by functions that
;;; directly call foreign functions from gtk_ or gdk_.


(defmacro until (condition &body body)
  `(do () (,condition) ,@body))

(defmacro while (condition &body body)
  `(until (not ,condition) ,@body))

(defun round-coordinate (x)
  ;; Das machen wir einfach mal wie in CLIM-CLX.
  (floor (+ x .5)))


;;;; GTKAIRO-PORT

(defclass gtkairo-pointer (standard-pointer)
  ((port :initarg :port :accessor port)
   (cursor :accessor pointer-cursor :initform :upper-left)))

(defclass gtkairo-port (basic-port)
  ((pointer :accessor port-pointer)
   (events-head :accessor events-head)
   (events-tail :accessor events-tail)
   (widgets->sheets :initform (make-hash-table) :accessor widgets->sheets)
   (metrik-medium :accessor metrik-medium)))

;;;(defmethod print-object ((object gtkairo-port) stream)
;;;  (print-unreadable-object (object stream :identity t :type t)
;;;    (format stream "~S ~S" :id (slot-value object 'id))))

(defun parse-gtkairo-server-path (path)
  path)

;;; FIXME: if :port-type and :server-path-parser aren't CLIM-specified
;;; keywords, they should be altered to be in some mcclim-internal
;;; package instead.
(setf (get :gtkairo :port-type) 'gtkairo-port)
(setf (get :gtkairo :server-path-parser) 'parse-gtkairo-server-path)

(defmethod initialize-instance :after ((port gtkairo-port) &rest initargs)
  (declare (ignore initargs))
  (setf (events-head port) (list nil))
  (setf (events-tail port) (events-head port))
  (setf (port-pointer port) (make-instance 'gtkairo-pointer :port port))
  ;; FIXME: it seems bizarre for this to be necessary
  (push (make-instance 'gtkairo-frame-manager :port port)
	(slot-value port 'climi::frame-managers))
  (when (zerop *g-threads-got-initialized*)
    (g_thread_init (cffi:null-pointer))
    (gdk_threads_init))
  (with-gtk ()
    ;; FIXME: hier koennten wir mindestens ein anderes --display uebergeben
    ;; wenn wir wollten
    (gtk_init 0 (cffi:null-pointer))
    (let ((cr (gdk_cairo_create
		(gdk_screen_get_root_window (gdk_screen_get_default)))))
      (setf (metrik-medium port)
            (make-instance 'metrik-medium :port port :cr cr))))
  (when clim-sys:*multiprocessing-p*
    (start-event-thread port)))

(defmethod destroy-port :before ((port gtkairo-port))
  )

(defun start-event-thread (port)
  (setf (climi::port-event-process port)
	(clim-sys:make-process
	 (lambda ()
	   (loop
	     (with-simple-restart
		 (restart-event-loop "Restart CLIM's event loop.")
	       (loop
		 (process-next-event port)))))
	 :name (format nil "~S's event process." port))))


;;;; Mirrors

(defclass mirror ()
    ((mediums :initform '() :accessor mirror-mediums)))

(defclass widget-mirror (mirror)
    ((widget :initarg :widget :accessor mirror-widget)
     (mediums :initform '() :accessor mirror-mediums)))

(defclass window-mirror (widget-mirror)
    ((window :initarg :window :accessor mirror-window)))

(defclass native-widget-mirror (widget-mirror)
    ((fixed :initarg :fixed :accessor mirror-fixed)))

(defclass drawable-mirror (mirror)
    ((drawable :initarg :drawable :accessor mirror-drawable)
     (mediums :initform '() :accessor mirror-mediums)))

(defmethod mirror-drawable ((mirror widget-mirror))
  (gtkwidget-gdkwindow (mirror-widget mirror)))

(defun widget->sheet (widget port)
  (gethash (cffi:pointer-address widget) (widgets->sheets port)))

(defun (setf widget->sheet) (newval widget port)
  (let ((address (cffi:pointer-address widget))
	(table (widgets->sheets port)))
    (if newval
	(setf (gethash address table) newval)
	(remhash address table))))

;;; Pixmaps: The spec doesn't say what a pixmap is, but the McCLIM
;;; frontend wants us to use its PIXMAP class.  So every backend
;;; copy&pastes PORT-(DE)ALLOCATE-PIXMAP to conform.  Why doesn't BASIC-PORT
;;; do this for us then?
(defmethod port-allocate-pixmap ((port gtkairo-port) sheet width height)
  (let ((pixmap (make-instance 'gtkairo-pixmap
		  :sheet sheet
		  :width width
		  :height height
		  :port port)))
    (when (sheet-grafted-p sheet)
      (realize-mirror port pixmap))
    pixmap))

(defmethod port-deallocate-pixmap ((port gtkairo-port) pixmap)
  (when (climi::port-lookup-mirror port pixmap)
    (destroy-mirror port pixmap)))


;;;; REALIZE-MIRROR

(defmethod realize-window-mirror (port sheet type)
  (with-gtk ()
    (let* ((q (compose-space sheet))
	   (window (gtk_window_new type))
	   (widget (gtk_fixed_new))
	   (width (round-coordinate (space-requirement-width q)))
	   (height (round-coordinate (space-requirement-height q)))
	   (mirror (make-instance 'window-mirror :window window :widget widget)))
      (gtk_window_set_title window (frame-pretty-name (pane-frame sheet)))
      (setf (widget->sheet widget port) sheet)
      (setf (widget->sheet window port) sheet)
      (connect-signals widget)
      (connect-window-signals window)
      (gtk_widget_set_double_buffered widget 0)
      (gtk_window_set_default_size window width height)
      (gtk_container_add window widget)
      (climi::port-register-mirror (port sheet) sheet mirror)
      mirror)))

(defmethod realize-mirror
    ((port gtkairo-port) (sheet climi::top-level-sheet-pane))
  (let ((mirror (realize-window-mirror port sheet GTK_WINDOW_TOPLEVEL)))
    (port-enable-sheet port sheet)
    mirror))

(defmethod realize-mirror
    ((port gtkairo-port) (sheet climi::unmanaged-top-level-sheet-pane))
  (realize-window-mirror port sheet GTK_WINDOW_POPUP))

;; copy&paste from port.lisp|CLX:
(defun sheet-desired-color (sheet)
  (multiple-value-bind (r g b)
      (color-rgb
       (typecase sheet
	 (sheet-with-medium-mixin
	   (medium-background sheet))
	 (basic-pane
	   ;; CHECKME [is this sensible?] seems to be
	   (let ((background (pane-background sheet)))
	     (if (typep background 'color)
		 background
		 +white+)))
	 (t
	   +white+)))
    (values (min (truncate (* r 65536)) 65535)
	    (min (truncate (* g 65536)) 65535)
	    (min (truncate (* b 65536)) 65535))))

(defmethod realize-mirror ((port gtkairo-port) (sheet mirrored-sheet-mixin))
  (with-gtk ()
    (let* ((parent (sheet-mirror (sheet-parent sheet)))
	   (q (compose-space sheet))
	   (widget (gtk_fixed_new))
	   (width (round-coordinate (space-requirement-width q)))
	   (height (round-coordinate (space-requirement-height q)))
	   (mirror (make-instance 'widget-mirror :widget widget)))
      (setf (widget->sheet widget port) sheet)
      ;; Das machen wir uns mal einfach und geben jedem Widget sein eigenes
      ;; Fenster, dann haben wir naemlich das Koordinatensystem und Clipping
      ;; wie aus CLX bekannt.  An und fuer sich koennten wir das aber auch
      ;; selbst per Cairo erledigen.  Muss man sich langfristig mal ueberlegen.
      (gtk_fixed_set_has_window widget 1)
      (gtk_widget_set_double_buffered widget 0)
      (connect-signals widget)
      (gtk_widget_set_size_request widget width height)
      (multiple-value-bind (x y)
	  (transform-position (climi::%sheet-mirror-transformation sheet) 0 0)
	(setf x (round-coordinate x))
	(setf y (round-coordinate y))
	(gtk_fixed_put (mirror-widget parent) widget x y))
      (climi::port-register-mirror (port sheet) sheet mirror)
      (cffi:with-foreign-object (color 'gdkcolor)
	(setf (cffi:foreign-slot-value color 'gdkcolor 'pixel) 0)
	(setf (values (cffi:foreign-slot-value color 'gdkcolor 'r)
		      (cffi:foreign-slot-value color 'gdkcolor 'g)
		      (cffi:foreign-slot-value color 'gdkcolor 'b))
	      (sheet-desired-color sheet))
	(gtk_widget_modify_bg widget 0 color))
      (when (sheet-enabled-p sheet)
	(gtk_widget_show widget))
      mirror)))

(defclass native-widget-mixin () ())

(defmethod realize-mirror ((port gtkairo-port) (sheet native-widget-mixin))
  (with-gtk ()
    (let* ((parent (sheet-mirror (sheet-parent sheet)))
	   (q (compose-space sheet))
	   (fixed (gtk_fixed_new))
	   (widget (realize-native-widget sheet))
	   (width (round-coordinate (space-requirement-width q)))
	   (height (round-coordinate (space-requirement-height q)))
	   (mirror
	    (make-instance 'native-widget-mirror :fixed fixed :widget widget)))
      (setf (widget->sheet fixed port) sheet)
      (setf (widget->sheet widget port) sheet)
      (gtk_fixed_set_has_window fixed 1)
      (connect-native-signals sheet widget)
      (gtk_widget_set_size_request fixed width height)
      (gtk_widget_set_size_request widget width height)
      (multiple-value-bind (x y)
	  (transform-position (climi::%sheet-mirror-transformation sheet) 0 0)
	(setf x (round-coordinate x))
	(setf y (round-coordinate y))
	(gtk_fixed_put (mirror-widget parent) fixed x y))
      (gtk_fixed_put fixed widget 0 0)
      (climi::port-register-mirror (port sheet) sheet mirror)
      (when (sheet-enabled-p sheet)
	(gtk_widget_show_all fixed))
      mirror)))

(defmethod realize-mirror ((port gtkairo-port) (pixmap-sheet climi::pixmap))
  (unless (climi::port-lookup-mirror port pixmap-sheet)
    (let* ((drawable
	    (mirror-drawable
	     (sheet-direct-mirror (climi::pixmap-sheet pixmap-sheet))))
	   (pixmap
	    (gdk_pixmap_new drawable
			    (round (pixmap-width pixmap-sheet))
			    (round (pixmap-height pixmap-sheet))
			    -1))
	   (mirror (make-instance 'drawable-mirror :drawable pixmap)))
      (climi::port-register-mirror port pixmap-sheet mirror)
      mirror)))


;;;; DESTROY-MIRROR

(defun destroy-window-mirror (port sheet)
  (with-gtk ()
    (let ((mirror (climi::port-lookup-mirror port sheet)))
      (destroy-mediums mirror)
      (gtk_widget_destroy (mirror-window mirror))
      (gtk-main-iteration port)
      (climi::port-unregister-mirror port sheet mirror)
      (setf (widget->sheet (mirror-widget mirror) port) nil))))

(defun destroy-mediums (mirror)
  (dolist (medium (mirror-mediums mirror))
    (when (cr medium)
      (cairo_destroy (cr medium))
      (setf (cr medium) nil)))
  (setf (mirror-mediums mirror) '()))

(defmethod destroy-mirror
    ((port gtkairo-port) (sheet climi::top-level-sheet-pane))
  (destroy-window-mirror port sheet))

(defmethod destroy-mirror
    ((port gtkairo-port) (sheet climi::unmanaged-top-level-sheet-pane))
  (destroy-window-mirror port sheet))

(defmethod destroy-mirror ((port gtkairo-port) (sheet mirrored-sheet-mixin))
  (with-gtk ()
    (let ((mirror (climi::port-lookup-mirror port sheet)))
      (destroy-mediums mirror)
      (gtk_widget_destroy (mirror-widget mirror))
      (climi::port-unregister-mirror port sheet mirror)
      (setf (widget->sheet (mirror-widget mirror) port) nil))))

(defmethod destroy-mirror ((port gtkairo-port) (pixmap-sheet climi::pixmap))
  (with-gtk ()
    (let ((mirror (climi::port-lookup-mirror port pixmap-sheet)))
      (when mirror
	(destroy-mediums mirror)
	(gdk_drawable_unref (mirror-drawable mirror))
	(climi::port-unregister-mirror port pixmap-sheet mirror)))))


;;;; Positioning and resizing

(defun reset-mediums (mirror)
  (dolist (medium (mirror-mediums mirror))
    (setf (cr medium) nil)))

(defmethod port-set-mirror-region
    ((port gtkairo-port) (mirror window-mirror) mirror-region)
  (with-gtk ()
    (reset-mediums mirror)
    (gtk_window_resize (mirror-window mirror)
		       (floor (bounding-rectangle-max-x mirror-region))
		       (floor (bounding-rectangle-max-y mirror-region)))
    ;; Nanu, ohne die Geometrie hier zu korrigieren kann das Fenster nur
    ;; vergroessert, nicht aber wieder verkleinert werden.
    (cffi:with-foreign-object (geometry 'gdkgeometry)
      (setf (cffi:foreign-slot-value geometry 'gdkgeometry 'min_width) 1)
      (setf (cffi:foreign-slot-value geometry 'gdkgeometry 'min_height) 1)
      (gtk_window_set_geometry_hints (mirror-window mirror)
				     (mirror-window mirror)
				     geometry
				     2))))

(defmethod port-set-mirror-region
    ((port gtkairo-port) (mirror mirror) mirror-region)
  (with-gtk ()
    (reset-mediums mirror)
    (gtk_widget_set_size_request
     (mirror-widget mirror)
     (floor (bounding-rectangle-max-x mirror-region))
     (floor (bounding-rectangle-max-y mirror-region)))))

(defmethod port-set-mirror-region
    ((port gtkairo-port) (mirror native-widget-mirror) mirror-region)
  (with-gtk ()
    (let ((w (floor (bounding-rectangle-max-x mirror-region)))
	  (h (floor (bounding-rectangle-max-y mirror-region))))
      (gtk_widget_set_size_request (mirror-fixed mirror) w h)
      (gtk_widget_set_size_request (mirror-widget mirror) w h))))

(defmethod port-set-mirror-transformation
    ((port gtkairo-port) (mirror window-mirror) mirror-transformation)
  (with-gtk ()
    (multiple-value-bind (x y)
	(transform-position mirror-transformation 0 0)
      (gtk_window_move (mirror-window mirror) (floor x) (floor y)))))

(defmethod port-set-mirror-transformation
    ((port gtkairo-port) (mirror mirror) mirror-transformation)
  (with-gtk ()
    (let* ((w (mirror-widget mirror))
	   (parent (cffi:foreign-slot-value w 'gtkwidget 'parent)))
      (multiple-value-bind (x y)
	  (transform-position mirror-transformation 0 0)
	(gtk_fixed_move parent w (floor x) (floor y))))))

(defmethod port-set-mirror-transformation
    ((port gtkairo-port) (mirror native-widget-mirror) mirror-transformation)
  (with-gtk ()
    (let* ((w (mirror-fixed mirror))
	   (parent (cffi:foreign-slot-value w 'gtkwidget 'parent)))
      (multiple-value-bind (x y)
	  (transform-position mirror-transformation 0 0)
	(gtk_fixed_move parent w (floor x) (floor y))))))


;;;; An und aus

(defmethod port-enable-sheet
    ((port gtkairo-port) (sheet mirrored-sheet-mixin))
  (with-gtk ()
    (let ((mirror (climi::port-lookup-mirror port sheet)))
      (gtk_widget_show (mirror-widget mirror)))))

(defmethod port-enable-sheet
    ((port gtkairo-port) (sheet native-widget-mixin))
  (with-gtk ()
    (let ((mirror (climi::port-lookup-mirror port sheet)))
      (gtk_widget_show_all (mirror-fixed mirror)))))

(defmethod port-enable-sheet
    ((port gtkairo-port) (sheet climi::top-level-sheet-pane))
  (with-gtk ()
    (let ((mirror (climi::port-lookup-mirror port sheet)))
      (gtk_widget_show (mirror-window mirror))
      (gtk_widget_show (mirror-widget mirror)))))

(defmethod port-enable-sheet
    ((port gtkairo-port) (sheet climi::unmanaged-top-level-sheet-pane))
  (with-gtk ()
    (let ((mirror (climi::port-lookup-mirror port sheet)))
      (gtk_widget_show (mirror-window mirror))
      (gtk_widget_show (mirror-widget mirror)))))

(defmethod port-disable-sheet
    ((port gtkairo-port) (sheet mirrored-sheet-mixin))
  (with-gtk ()
    (let ((mirror (climi::port-lookup-mirror port sheet)))
      (gtk_widget_hide (mirror-widget mirror)))))

(defmethod port-disable-sheet
    ((port gtkairo-port) (sheet native-widget-mixin))
  (with-gtk ()
    (let ((mirror (climi::port-lookup-mirror port sheet)))
      (gtk_widget_hide_all (mirror-fixed mirror)))))

(defmethod port-disable-sheet
    ((port gtkairo-port) (sheet climi::top-level-sheet-pane))
  (with-gtk ()
    (let ((mirror (climi::port-lookup-mirror port sheet)))
      (gtk_widget_hide (mirror-window mirror)))))


;;;; JUNK

(defmethod mirror-transformation ((port gtkairo-port) mirror)
  ())

;; die sind verkehrt, die gibt's gar nicht
;;;(defmethod port-set-sheet-transformation
;;;    ((port gtkairo-port) (graft graft) transformation)
;;;  ())
;;;
;;;(defmethod port-set-sheet-transformation
;;;    ((port gtkairo-port) (sheet mirrored-sheet-mixin) transformation)
;;;  ())


;;;; Vermischtes

(defmethod port-motion-hints
    ((port gtkairo-port) (sheet mirrored-sheet-mixin))
  (with-gtk ()
    (logtest GDK_POINTER_MOTION_HINT_MASK
	     (gtk_widget_get_events
	      (mirror-widget (sheet-direct-mirror sheet))))))

(defmethod (setf port-motion-hints)
    (value (port gtkairo-port) (sheet mirrored-sheet-mixin))
  (with-gtk ()
    (let* ((widget (mirror-widget (sheet-direct-mirror sheet)))
	   (oldval (gtk_widget_get_events widget))
	   (newval
	    (if value
		(logior oldval GDK_POINTER_MOTION_HINT_MASK)
		(logandc2 oldval GDK_POINTER_MOTION_HINT_MASK))))
      (gtk_widget_set_events widget newval))))

(defmethod make-graft
    ((port gtkairo-port) &key (orientation :default) (units :device))
  (make-instance 'gtkairo-graft
                 :port port :mirror (gensym)
                 :orientation orientation :units units))

(defmethod text-style-mapping
    ((port gtkairo-port) text-style &optional character-set)
  (error "text-style-mapping called, what now?"))

(defmethod (setf text-style-mapping)
    (font-name (port gtkairo-port)
     (text-style text-style) &optional character-set)
  (error "(setf text-style-mapping) called, what now?"))

(defmethod port-character-width ((port gtkairo-port) text-style char)
  (error "port-character-width called, what now?"))

(defmethod port-string-width
    ((port gtkairo-port) text-style string &key (start 0) end)
  (error "port-string-width called, what now?"))

(defmethod port-mirror-width ((port gtkairo-port) sheet)
  (error "port-mirror-width called, we thought the frontend doesn't do that"))

(defmethod port-mirror-height ((port gtkairo-port) sheet)
  (error "port-mirror-height called, we thought the frontend doesn't do that"))

(defmethod graft ((port gtkairo-port))
  (first (climi::port-grafts port)))

(defun %gdk-display-get-pointer ()
  (with-gtk ()
    (cffi:with-foreign-object (x :int)
      (cffi:with-foreign-object (y :int)
	(cffi:with-foreign-object (mask :int)
	  (gdk_display_get_pointer (gdk_display_get_default)
				   (cffi:null-pointer) ;FIXME: screen
				   x
				   y
				   mask)
	  (values
	   (cffi:mem-aref x :int)
	   (cffi:mem-aref y :int)
	   (cffi:mem-aref mask :int)))))))

(defun mirror-pointer-position (mirror)
  (with-gtk ()
    (cffi:with-foreign-object (x :int)
      (cffi:with-foreign-object (y :int)
	(gtk_widget_get_pointer (mirror-widget mirror) x y)
	(values
	 (cffi:mem-aref x :int)
	 (cffi:mem-aref y :int))))))

(defmethod pointer-position ((pointer gtkairo-pointer))
  (%gdk-display-get-pointer))

(defmethod pointer-button-state ((pointer gtkairo-pointer))
  (with-gtk ()
    (gdkmodifiertype->all-buttons (nth-value 2 (%gdk-display-get-pointer)))))

(defmethod port-modifier-state ((port gtkairo-port))
  (with-gtk ()
    (gdkmodifiertype->modifier-state (nth-value 2 (%gdk-display-get-pointer)))))

;;; port.lisp|CLX says:
;;;   XXX Should we rely on port-pointer-sheet being correct? -- moore
(defmethod synthesize-pointer-motion-event ((pointer gtkairo-pointer))
  (with-gtk ()
    (let* ((port (port pointer))
	   (sheet (climi::port-pointer-sheet port)))
      (when sheet
	(let ((mirror (sheet-direct-mirror sheet)))
	  (when mirror
	    (multiple-value-bind (root-x root-y state)
		(%gdk-display-get-pointer)
	      ;; FIXME ;when same-screen-p
	      (multiple-value-bind (x y)
		  (mirror-pointer-position mirror)
		(make-instance 'pointer-motion-event
		  :pointer 0
		  :button (gdkmodifiertype->one-button state)
		  :x x
		  :y y
		  :graft-x root-x
		  :graft-y root-y
		  :sheet sheet
		  :modifier-state (gdkmodifiertype->modifier-state state)
		  ;; The event initialization code will give us a
		  ;; reasonable timestamp.
		  :timestamp 0)))))))))

(defmethod %set-port-keyboard-focus ((port gtkairo-port) focus &key timestamp)
  (declare (ignore timestamp))
  (with-gtk ()
    (gtk_widget_grab_focus (mirror-widget (sheet-mirror focus)))))

(defmethod port-force-output ((port gtkairo-port))
  (with-gtk ()
    (gdk_display_flush (gdk_display_get_default))))

;; FIXME: What happens when CLIM code calls tracking-pointer recursively?
(defmethod port-grab-pointer ((port gtkairo-port) pointer sheet)
  (with-gtk ()
    (let* ((gtkwidget (mirror-widget (sheet-direct-mirror sheet)))
	   (status
	    (gdk_pointer_grab (gtkwidget-gdkwindow gtkwidget)
			      0
			      (logior GDK_POINTER_MOTION_MASK
				      GDK_BUTTON_PRESS_MASK
				      GDK_BUTTON_RELEASE_MASK)
			      (cffi:null-pointer)
			      (cffi:null-pointer)
			      GDK_CURRENT_TIME)))
      ;; emergency ungrab:
;;;      (sb-thread:make-thread (lambda ()
;;;			       (sleep 10)
;;;			       (tr :timeout!)
;;;			       (with-gtk ()
;;;				 (gdk_pointer_ungrab GDK_CURRENT_TIME))))
      (zerop status))))

(defmethod port-ungrab-pointer ((port gtkairo-port) pointer sheet)
  (declare (ignore pointer sheet))
  (with-gtk ()
    (gdk_pointer_ungrab GDK_CURRENT_TIME)))

(defmethod distribute-event :around ((port gtkairo-port) event)
  (call-next-method))

(defmethod set-sheet-pointer-cursor ((port gtkairo-port) sheet cursor)
  ())        

(defmethod bind-selection ((port gtkairo-port) window &optional time)
  ())

(defmethod release-selection ((port gtkairo-port) &optional time)
  ())

(defmethod request-selection ((port gtkairo-port) requestor time)
  ())

(defmethod get-selection-from-event ((port gtkairo-port) event)
  ())

(defmethod send-selection ((port gtkairo-port) event string)
  nil)
